from enum import Enum
from typing import Optional

from .camera import Camera
from .image import ImageSize


class ImageRotation(str, Enum):
    NONE: int = 0
    RIGHT: int = 90
    UPSIDE_DOWN: int = 180
    LEFT: int = 270


class UsbCamera(Camera):
    capture: bool = True
    detect: bool = False
    connected: bool = False
    resolution: Optional[ImageSize] = None
    exposure: Optional[float] = None
    color: Optional[str] = None
    rotation: ImageRotation = ImageRotation.NONE

    @property
    def image_resolution(self) -> Optional[ImageSize]:
        if self.rotation == ImageRotation.LEFT or self.rotation == ImageRotation.RIGHT:
            return ImageSize(width=self.resolution.height, height=self.resolution.width)
        else:
            return self.resolution.copy()



from .binary_renderer import BinaryRenderer
from .distance_map import DistanceMap
from .experiments import generate_experiment
from .grid import Grid
from .obstacle_map import ObstacleMap
from .planner_process import PlannerCommand, PlannerProcess, PlannerResponse, PlannerState, \
    PlannerGetStateCommand, PlannerGrowMapCommand, PlannerSearchCommand, PlannerTestCommand
from .robot_renderer import RobotRenderer
from .steps import Path, Step
from . import plot_tools



import io
import random
import time

import PIL as pil
import rosys

from .. import event
from ..world import Image, UsbCamera
from .actor import Actor


class UsbCameraSimulator(Actor):
    interval: float = 1

    async def step(self):
        await super().step()

        for camera in self.world.usb_cameras.values():
            size = camera.calibration.intrinsics.size
            image = Image(time=self.world.time, camera_id=camera.id, size=size)
            if rosys.is_test:
                image.data = b'test data'
            else:
                image.data = await rosys.run.cpu_bound(self.create_image_data, camera)
            camera.images.append(image)

    async def create(
        self, uid: str = '',
        x: float = 0, y: float = 0, z: float = 1,
        yaw: float = 0, tilt_x: float = 0, tilt_y: float = 0,
        color=None,
    ):
        if uid == '':
            uid = f'simulated_cam_{len(self.world.usb_cameras)}'
        camera = UsbCamera(id=uid)
        camera.color = color or '#' + ('%06x' % random.randint(0, 0xFFFFFF))
        camera.set_perfect_calibration(x, y, z, yaw, tilt_x, tilt_y)
        self.world.usb_cameras[uid] = camera
        await event.call(event.Id.NEW_CAMERA, self.world.usb_cameras[uid])

    @staticmethod
    def create_image_data(camera: UsbCamera):
        size = camera.calibration.intrinsics.size
        img = pil.Image.new('RGB', size=(size.width, size.height), color=camera.color)
        d = pil.ImageDraw.Draw(img)
        text = f'{camera.id}: {time.time()}'
        d.text((img.width/2-len(text)*3, img.height/2-5), text, fill=(255, 255, 255))
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG')
        return img_byte_arr.getvalue()



import re
import shutil
from dataclasses import dataclass
from typing import Any, Optional

import cv2
import rosys

from .. import event
from ..world import Image, ImageSize, UsbCamera
from .actor import Actor


@dataclass
class Device:
    '''devices specific infos are kept seperately (world should not be aware of them)'''
    uid: str
    video_id: int
    capture: Any  # cv2.VideoCapture device
    resolution: Optional[ImageSize] = None
    exposure: Optional[float] = None


def process_image(image, rotation: rosys.world.ImageRotation) -> bytes:
    if rotation == rosys.world.ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == rosys.world.ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == rosys.world.ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return cv2.imencode('.jpg', image)[1].tobytes()


class UsbCameraCapture(Actor):
    interval: float = 0.3

    def __init__(self):
        super().__init__()
        self.devices: dict[str, Device] = {}
        self.last_scan: Optional[float] = None

    async def step(self):
        await super().step()
        await self.update_device_list()
        for uid, camera in self.world.usb_cameras.items():
            if not camera.capture or not camera.connected:
                continue
            try:
                image = await rosys.run.io_bound(self.capture_image, uid)
                if image is None:
                    self.disconnect(uid)
                    continue
                bytes = await rosys.run.cpu_bound(process_image, image, camera.rotation)
                size = camera.resolution or ImageSize(width=800, height=600)
                camera.images.append(Image(camera_id=uid, data=bytes, time=self.world.time, size=size))
            except:
                self.log.exception(
                    f'could not capture image from {uid}; disconnecting device /dev/video{self.devices[uid].video_id}')
                self.disconnect(uid)
        self.purge_old_images()

    def capture_image(self, id) -> Any:
        capture = self.devices[id].capture
        _, image = capture.read()
        return image

    def disconnect(self, uid: str):
        self.log.info(f'disconnecting {uid}')
        camera = self.world.usb_cameras[uid]
        camera.connected = False
        self.devices[camera.id].capture.release()
        del self.devices[camera.id]

    def purge_old_images(self):
        for camera in self.world.usb_cameras.values():
            while camera.images and camera.images[0].time < self.world.time - 5 * 60.0:
                del camera.images[0]

    async def update_device_list(self):
        if self.last_scan is not None and self.world.time < self.last_scan + 30:  # scan every 30 sec
            return
        self.last_scan = self.world.time
        output = await rosys.run.sh(['v4l2-ctl', '--list-devices'])
        for c in self.world.usb_cameras.values():
            c.connected = False
        for infos in output.split('\n\n'):
            if 'Cannot open device' in infos:
                continue
            match = re.search('\((.*)\)', infos)
            if match is None:
                continue
            uid = match.group(1)
            if uid not in self.world.usb_cameras:
                self.world.usb_cameras[uid] = UsbCamera(id=uid)
                self.log.info(f'adding camera {uid}')
                await event.call(event.Id.NEW_CAMERA, self.world.usb_cameras[uid])
            lines = infos.splitlines()
            if 'dev/video' not in lines[1]:
                continue
            num = int(lines[1].strip().lstrip('/dev/video'))
            if uid not in self.devices:
                capture = self.get_capture_device(num)
                if capture is None:
                    if uid in self.world.usb_cameras:
                        self.world.usb_cameras[uid].connected = False
                else:
                    self.devices[uid] = Device(uid, num, capture)
            if uid in self.devices:
                self.world.usb_cameras[uid].connected = True
                await self.update_parameters(self.devices[uid])

    def get_capture_device(self, index: int):
        try:
            capture = cv2.VideoCapture(index)
            if capture is None:
                self.log.error(f'video device {index} is unavailable')
            elif not capture.isOpened():
                self.log.error(f'video device {index} can not be opened')
                capture.release()
            else:
                return capture
        except:
            self.log.exception(f'{index} device failed')

    async def tear_down(self):
        await super().tear_down()
        for camera in self.world.usb_cameras.values():
            camera.connected = False
        [device.capture.release() for device in self.devices.values()]
        self.devices.clear()

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('v4l2-ctl') is not None

    async def update_parameters(self, device: Device):
        output = await self.run_v4l(device, '--all')
        size = re.search('Width/Height.*: (\d*)/(\d*)', output)
        device.resolution = ImageSize(width=int(size.group(1)), height=int(size.group(2)))
        camera = self.world.usb_cameras[device.uid]
        if camera.resolution and camera.resolution != device.resolution:
            self.log.info(f'updating resolution of {camera.id} from {device.resolution} to {camera.resolution}')
            await rosys.run.io_bound(UsbCameraCapture.update_resolution, device, camera.resolution)
            # TODO read exposure from output and update it correctly
            #     if device.exposure != camera.brightness:
            #         await self.update_exposure(device)

            # async def update_exposure(self, device: Device):
            #     exposure = self.world.usb_cameras[device.uid].exposure
            #     # TODO if expousre is None, set auto exposure
            #     await self.run_v4l(device, '--set-ctrl=brightness=0', '--set-ctrl=exposure_auto=1', f'--set-ctrl=exposure_absolute={exposure}')
            #     self.log.info(f'using new exposure {exposure}')

    @staticmethod
    def update_resolution(device: Device, size: ImageSize):
        device.capture.set(cv2.CAP_PROP_FRAME_WIDTH, size.width)
        device.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, size.height)

    async def run_v4l(self, device: Device, *args):
        cmd = ['v4l2-ctl', '-d', str(device.video_id)]
        cmd.extend(args)
        return await rosys.run.sh(cmd)



from datetime import datetime, timedelta
from typing import Optional

import socketio

from .. import event, task_logger
from ..helpers import sleep
from ..world import BoxDetection, Detections, Image, PointDetection
from .actor import Actor


class Detector(Actor):
    interval: float = 1.0

    def __init__(self, port: int = 8004):
        super().__init__()
        self.sio = socketio.AsyncClient()
        self.is_detecting: bool = False
        self.next_image: Optional[Image] = None
        self.port = port

    @property
    def is_connected(self):
        return self.sio.connected

    async def step(self):
        if not self.is_connected and not await self.connect():
            await sleep(3.0)
            return

        await self.try_start_one_upload()

    async def connect(self) -> bool:
        try:
            url = f'ws://localhost:{self.port}'
            self.log.info(f'connecting to detector at {url}')
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3.0)
            self.log.info('connected successfully')
            return True
        except:
            self.log.exception('connection failed; trying again')
            return False

    async def try_start_one_upload(self):
        if datetime.now() < self.world.upload.last_upload + timedelta(minutes=self.world.upload.minimal_minutes_between_uploads):
            return

        upload_images = [image for image in self.world.upload.queue if image.data]
        if not upload_images:
            return

        self.world.upload.last_upload = datetime.now()
        self.world.upload.queue.clear()  # old images should not be uploaded later when the robot is inactive
        task_logger.create_task(self.upload(upload_images[0]), name='upload_image')

    async def upload(self, image: Image):
        try:
            await self.sio.emit('upload', {'image': image.data, 'mac': image.camera_id})
        except:
            self.log.exception(f'could not upload {image.id}')

    async def detect(self, image: Image) -> None:
        if not self.is_connected:
            return

        self.next_image = image
        if self.is_detecting:
            return

        while self.next_image is not None:
            try:
                image = self.next_image
                self.next_image = None
                self.is_detecting = True
                result = await self.sio.call('detect', {'image': image.data, 'mac': image.camera_id}, timeout=1)
                box_detections = [BoxDetection.parse_obj(d) for d in result.get('box_detections', [])]
                point_detections = [PointDetection.parse_obj(d) for d in result.get('point_detections', [])]
                image.detections = Detections(boxes=box_detections, points=point_detections)
            except:
                self.log.exception(f'could not detect {image.id}')
            else:
                event.emit(event.Id.NEW_DETECTIONS, image)
            finally:
                self.is_detecting = False

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'



import abc
from dataclasses import dataclass, field
import icecream
import heapq
import logging
from multiprocessing import Process
from multiprocessing.connection import Connection
import numpy as np
import time
from typing import Any, Optional
import uuid

from .distance_map import DistanceMap
from .grid import Grid
from .obstacle_map import Obstacle, ObstacleMap
from .steps import Step
from ...helpers import angle
from ...world import Area, PathSegment, Point, Pose, Spline


@dataclass
class PlannerState:
    robot_outline: list[tuple[float, float]]
    obstacle_map: Optional[ObstacleMap] = None
    small_obstacle_map: Optional[ObstacleMap] = None
    distance_map: Optional[DistanceMap] = None
    obstacles: list[Obstacle] = field(default_factory=list)
    areas: list[Area] = field(default_factory=list)
    goal: Optional[Pose] = None


@dataclass
class PlannerCommand(abc.ABC):
    deadline: float

    def __post_init__(self):
        self.id = str(uuid.uuid4())


@dataclass
class PlannerGetStateCommand(PlannerCommand):
    pass


@dataclass
class PlannerGrowMapCommand(PlannerCommand):
    points: list[Point]


@dataclass
class PlannerSearchCommand(PlannerCommand):
    areas: list[Area]
    obstacles: list[Obstacle]
    start: Pose
    goal: Pose
    backward: bool


@dataclass
class PlannerTestCommand(PlannerCommand):
    areas: list[Area]
    obstacles: list[Obstacle]
    spline: Spline
    backward: bool = False


@dataclass
class PlannerResponse:
    id: str
    deadline: float
    content: Any


class PlannerProcess(Process):

    def __init__(self, connection: Connection, robot_outline: list[tuple[float, float]]):
        super().__init__()
        self.log = logging.getLogger('rosys.pathplanning.PlannerProcess')
        self.connection = connection
        self.state = PlannerState(robot_outline)

    def run(self):
        icecream.install()  # NOTE provide ic(...) in subprocess
        while True:
            try:
                cmd = self.connection.recv()
            except EOFError:
                self.log.info('PlannerProcess stopped')
                return
            try:
                if isinstance(cmd, PlannerGetStateCommand):
                    self.respond(cmd, self.state)
                if isinstance(cmd, PlannerGrowMapCommand):
                    self.grow_obstacle_map(cmd.points, cmd.deadline)
                    self.respond(cmd, None)
                if isinstance(cmd, PlannerSearchCommand):
                    additional_points = [cmd.start.point, cmd.goal.point]
                    self.update_obstacle_map(cmd.areas, cmd.obstacles, additional_points, cmd.deadline)
                    self.update_distance_map(cmd.goal, cmd.deadline)
                    self.respond(cmd, self.search(cmd.goal, cmd.start, cmd.backward, cmd.deadline))
                if isinstance(cmd, PlannerTestCommand):
                    self.update_obstacle_map(cmd.areas, cmd.obstacles, [cmd.spline.start, cmd.spline.end], cmd.deadline)
                    self.respond(cmd, bool(self.state.obstacle_map.test_spline(cmd.spline, cmd.backward)))
            except Exception as e:
                self.log.exception(f'failed to compute cmd "{cmd}"')
                self.respond(cmd, e)

    def respond(self, cmd: PlannerCommand, content: Any):
        self.connection.send(PlannerResponse(cmd.id, cmd.deadline, content))

    def recreate_obstacle_map(self,
                              areas: list[Area],
                              obstacles: list[Obstacle],
                              additional_points: list[Point],
                              deadline: float):
        points = [p for obstacle in obstacles for p in obstacle.outline]
        points += [p for area in areas for p in area.outline]
        points += additional_points
        g1 = Grid.from_points(points, 0.1, 36, padding=1.0)
        g2 = Grid.from_points(points, 0.2, 0, padding=1.0)
        self.state.obstacle_map = ObstacleMap.from_world(self.state.robot_outline, areas, obstacles, g1, deadline)
        self.state.small_obstacle_map = ObstacleMap.from_world(self.state.robot_outline, areas, obstacles, g2, deadline)

    def update_obstacle_map(self,
                            areas: list[Area],
                            obstacles: list[Obstacle],
                            additional_points: list[Point],
                            deadline: float):
        if self.state.obstacle_map and \
                self.state.areas == areas and \
                self.state.obstacles == obstacles and \
                all(self.state.obstacle_map.grid.contains(point, padding=1.0) for point in additional_points):
            return
        self.recreate_obstacle_map(areas, obstacles, additional_points, deadline)
        self.state.distance_map = None
        self.state.areas = areas
        self.state.obstacles = obstacles

    def grow_obstacle_map(self, points: list[Point], deadline: float):
        if self.state.obstacle_map is not None and \
                all(self.state.obstacle_map.grid.contains(point, padding=1.0) for point in points):
            return
        if self.state.obstacle_map is not None:
            bbox = self.state.obstacle_map.grid.bbox
            points.append(Point(x=bbox[0],         y=bbox[1]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]))
            points.append(Point(x=bbox[0],         y=bbox[1]+bbox[3]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]+bbox[3]))
        self.recreate_obstacle_map(self.state.areas, self.state.obstacles, points, deadline)
        self.state.distance_map = None

    def update_distance_map(self, goal: Pose, deadline: float) -> None:
        if self.state.distance_map and self.state.goal == goal:
            return
        self.state.distance_map = DistanceMap(self.state.small_obstacle_map, goal, deadline)
        self.state.goal = goal

    def search(self, goal: Pose, start: Pose, backward: bool, deadline: float) -> Optional[list[PathSegment]]:
        step_dist = 0.5
        num_candidates = 16

        pose = (start.x, start.y, start.yaw)
        heap = [(self.state.distance_map.interpolate(start.x, start.y), 0, Step(pose))]
        visited = set()

        if self.state.obstacle_map.test(goal.x, goal.y, goal.yaw):
            raise RuntimeError('target pose is unreachable')

        while pose != (goal.x, goal.y, goal.yaw):
            if time.time() > deadline:
                raise TimeoutError('could not find path')

            try:
                _, travel_cost, step = heapq.heappop(heap)
            except IndexError:
                raise RuntimeError('could not find path')

            pose = step.target
            tup = tuple(np.round(pose, 3))
            if tup in visited:
                continue
            visited.add(tup)

            if pose == (goal.x, goal.y, goal.yaw):
                break

            goal_dist = np.sqrt((goal.x - pose[0])**2 + (goal.y - pose[1])**2)
            goal_candidate = [(goal.x, goal.y, goal.yaw)] if goal_dist < 3 * step_dist else []
            fw_candidates = [
                (pose[0] + step_dist * np.cos(yaw), pose[1] + step_dist * np.sin(yaw), yaw)
                for yaw in np.linspace(-np.pi / 2, np.pi / 2, num_candidates // 2 + 1)[1:-1] + pose[2]
            ]
            bw_candidates = [
                (pose[0] - step_dist * np.cos(yaw), pose[1] - step_dist * np.sin(yaw), yaw)
                for yaw in np.linspace(-np.pi / 2, np.pi / 2, num_candidates // 2 + 1)[1:-1] + pose[2]
            ]

            for candidate in goal_candidate + fw_candidates + bw_candidates:
                tup = tuple(np.round(candidate, 3))
                if tup in visited:
                    continue

                if self.state.obstacle_map.test(*candidate):
                    continue

                costs = None
                for backward in [False, True]:
                    if backward and candidate in fw_candidates:
                        continue
                    if not backward and candidate in bw_candidates:
                        continue

                    next_step = Step(candidate, step, backward=backward)
                    if candidate in goal_candidate:
                        if not next_step.is_healthy():
                            continue

                    if self.state.obstacle_map.get_minimum_spline_distance(next_step.spline, backward=backward) < 0.1:
                        continue

                    if costs is None:
                        dist = self.state.distance_map.interpolate(candidate[0], candidate[1])[0]
                        new_travel_cost = travel_cost + step_dist
                        if step.backward != next_step.backward:
                            new_travel_cost += step_dist
                        yaw_cost = 0
                        if dist > step_dist:
                            gx, gy = self.state.distance_map.gradient(candidate[0], candidate[1])
                            target_yaw = np.arctan2(gy, gx) if backward else np.arctan2(-gy, -gx)
                            yaw_error = abs(angle(candidate[2], target_yaw[0]))
                            if yaw_error > np.pi / 2:
                                yaw_cost = 5 * yaw_error
                        obstacle_cost = max(1.0 - self.state.obstacle_map.get_distance(*candidate), 0.0)
                        remaining_cost = dist + 0.7 * new_travel_cost + yaw_cost + 0.3 * obstacle_cost
                        costs = (remaining_cost, new_travel_cost)

                    heapq.heappush(heap, (costs[0], costs[1], next_step))

        path = step.backtrace()
        path.smooth(self.state.obstacle_map, control_dist=0.5)
        return [PathSegment(spline=step.spline, backward=step.backward) for step in path[1:]]
